/*******************************************************************************
 *
 * buffer.c  
 *                   
 * Provides a data structure as well as supporting functions for loading 
 * a source file into memory and parsing it. 
 * 
 * Cheetah Web Browser
 * Copyright (C) 2001 Garett Spencley 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 *
 *******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "buffer.h"
#include "error.h"
#include "http.h"

#define INIT_SIZE 512

/*
 * create_empty_buffer() - creates an empty buffer
 * with default values
 */

buffer_t *create_empty_buffer()
{
	buffer_t *result;

	result = (buffer_t *)malloc(sizeof(buffer_t));
	if(!result)
		return NULL;

	result->lines        = NULL;

	result->num_lines    = 0;
	result->line_counter = 0;
	result->char_counter = 0;
	result->size         = 0;

	return result;
}

/*
 * load_file() - loads a file into a buffer and returns it
 */

buffer_t *buffer_load_file(char *file)
{
	FILE *infile;
	buffer_t *result;
	char buf[4096];

	if(!file)
		return NULL;

	result = create_empty_buffer();
	if(!result)
		return NULL;

	infile = fopen(file, "r");
	if(!infile) 
		return NULL;

	while(fgets(buf, 4095, infile))
		if(buffer_add_line(result, buf) < 0) 
			error("Out of memory.");

	fclose(infile);

	return result;
}

/*
 * http_get_document() - retrieves a document through http
 * and loads it into a buffer
 */

buffer_t *http_get_document(char *host, char *doc)
{
	buffer_t *buf;
	int result;

	result = http_connect(host);
	if(result < 0) {
		http_perror(host);
		return NULL; 
	}

	buf = http_get(doc);
	if(!buf) {
		http_perror(host);
		return NULL;
	}

	http_close();

	return buf;
}

/*
 * dispose_buffer() - disposes a buffer
 */

void dispose_buffer(buffer_t *buffer)
{
	free(buffer->lines);
	free(buffer);
}

/*
 * buffer_get_line() - returns the next line in the buffer
 */

void buffer_get_line(buffer_t *buf, char *dst, size_t n)
{
	if(!buf || !dst)
		return;

	if(!(buf->lines && buf->num_lines))
		dst[0] = '\0';
	else {
		strncpy(dst, buf->lines[buf->line_counter], n-1);
		dst[n-1] = '\0';
	}
}

/*
 * buffer_add_line() - adds a line onto the buffer
 */

int buffer_add_line(buffer_t *buf, const char *line)
{
	if(!buf || !line)
		return -1;

	/* Allocate our buffer if necessary */

	if(!buf->lines) {

		buf->lines = (char **)malloc(sizeof(char **) * INIT_SIZE);
		if(!buf->lines) {
			error("out of memory.");
			return -1;
		}
		buf->size = INIT_SIZE;
	}
	
	/* Resize our buffer if necessary */

	if(buf->num_lines == buf->size) {

		buf->lines = (char **)realloc(buf->lines, 
						sizeof(char **) * (buf->size + INIT_SIZE));
		if(!buf->lines) {
			error("out of memory.");
			return -1;
		}

		buf->size += INIT_SIZE;
	}

	/* Add the line */

	buf->lines[buf->num_lines] = strdup(line);
	if(!buf->lines[buf->num_lines]) {
		error("Out of memory.");
		return -1;
	}

	++buf->num_lines;

	return 0;
}

/*
 * buffer_get_char() - returns next char in a buffer
 */

int buffer_get_char(buffer_t *buf)
{
	int max_char;

	if(!buf)
		return -1;
	
	/* If we are pointing to the last char on a line
	   then return the first char on the next line */

	max_char = strlen(buf->lines[buf->line_counter]);

	if(buf->char_counter >= max_char) {

		++buf->line_counter;
		buf->char_counter = 0;

		/* If we're at the end of the buffer then return EOF */

		if(buf->line_counter >= buf->num_lines)
			return EOF;
	}

	/* Return the next char */

	return buf->lines[buf->line_counter][buf->char_counter++];
}

/*
 * buffer_unget_char() - puts a char back into the buffer
 */

void buffer_unget_char(int c, buffer_t *buf)
{
	/* If char_counter is 0 that means we need to step
	   back a line */

	if(buf->char_counter == 0)  {

		if(buf->line_counter == 0)
			return;	
		else {
			--buf->line_counter;
			buf->char_counter = strlen(buf->lines[buf->line_counter]);
		}
	} else 
		--buf->char_counter;

	/* Make sure that 'c' is really the char we are pointing
	   to now. */

	if(buf->lines[buf->line_counter][buf->char_counter] != c) {
		/* Do something */
	}
}

#ifdef _DEBUG
void print_buffer(buffer_t *buf)
{
	int i;

	for(i = 0; i < buf->num_lines; i++)
		fputs(buf->lines[i], stdout);	

	fputs("\n", stdout);
	fflush(stdout);
}
#endif
