/*******************************************************************************
 *
 * htmlparser.c - HTML parsing engine
 *
 * Cheetah Web Browser
 * Copyright (C) 2001 Garett Spencley 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 *
 *******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "error.h"
#include "htmlparser.h"
#include "htmltokenizer.h"
#include "string_t.h"
#include "buffer.h"
#include "html.h"

#define HTML_TAG_BUFFER_SIZE     16
#define HTML_CONTENT_BUFFER_SIZE 32

static unsigned short END_HTML_DOC;

buffer_t *source;

int html_parser_open(buffer_t *buffer);
void html_parser_close();

int find_next_tag();
char *extract_tag(char *tag);
html_tag *create_html_tag_struct(char *tag_string);
void free_html_tag(html_tag *tag);
void process_tag(html_tag *tag);

/* 
 * html_parser_init() - initializes the html parser
 */

int html_parser_open(buffer_t *buffer)
{
	if(!buffer) 
		return -1;

	source = buffer;

	return 0;
}

/*
 * html_parser_close() - closes the html parser 
 */

void html_parser_close()
{
	source = NULL;
}

/*
 * extract_tag()
 *
 * Returns only the tag/element within 'tag'
 */

char *extract_tag(char *tag)
{
	char *result, *p;

	if(!tag)
		return NULL;

	result = (char *)malloc(sizeof(char) * strlen(tag) + 1);
	if(!result)
		return NULL;

	p = result;

	while(*tag && !isspace(*tag))
		*p++ = *tag++;

	*p = 0;

	return result;
}

/*
 * extracts everything until the next tag is found.
 */

char *get_text(void)
{
	int c, counter;
	char *result, *tmp;
	int bufsize;

	/* Create a default buffer */

	bufsize = HTML_CONTENT_BUFFER_SIZE;

	result = (char *)malloc(sizeof(char) * bufsize);
	if(!result) 
		return NULL;

	counter = 0;

	/* Fill in the buffer char by char */

	while((c = buffer_get_char(source)) != '<') {

		if(c == EOF) {
			free(result);
			return NULL;
		}

		/* Resize the buffer if necessary */

		if(counter >= bufsize) {
			
			tmp = strdup(result);
			if(!tmp) {
				free(result);
				return NULL;
			}

			bufsize += HTML_CONTENT_BUFFER_SIZE;

			result = (char *)realloc(result, sizeof(char) * bufsize);
			if(!result) {
				free(tmp);
				return NULL;
			}	

			strcpy(result, tmp);
			free(tmp);
		}

		result[counter++] = c;
	}
	result[counter] = 0;

	return result;
}

/*
 * find_next_tag()
 *
 * Locates the next tag in the source buffer
 */

int find_next_tag(void)
{
	int c;

	while(1) {
			
		while((c = buffer_get_char(source)) != '<') {

			if(c == EOF)
				return -1;
		}

		/* If the tag is a comment handle it, otherwise
		 * we've found the next tag so bail out. */

		if((c = buffer_get_char(source)) != '!') {

			buffer_unget_char(c, source);
			break;

		} else {

			/* skip until end of the comment */

			while(1) {
					
				/* This just skips over the DTD. It's only temporary */

				if((c = buffer_get_char(source)) == 'D') {
					while((c = buffer_get_char(source)) != '>');
					break;
				}

				/* Not a DTD continue skipping until end of comment */

				while((c = buffer_get_char(source)) != '-');

				if((c = buffer_get_char(source)) != '-')
					continue;
				
				if((c = buffer_get_char(source)) != '>')
					continue;

				break;
			}
		}
	}	

	return 0;
}

/*
 * create_html_tag()
 *
 * creates an html_tag structure from a tag
 */

html_tag *create_html_tag_struct(char *tag_string)
{
	char *elem, *args;
	html_tag *result;
	string_t *string;

	/* Allocate an empty html tag */

	result = (html_tag *)malloc(sizeof(html_tag));
	if(!result) {
		error("Out of memory.");
		return NULL;
	}

	/* Create a string for parsing */

	string = create_string(tag_string);
	if(!string) {
		error("Out of memory.");
		return NULL;
	}

	/* Grab the tag element */

	elem = get_ident(string);
	if(!elem) {
		error("out of memory.");
		return NULL;
	}
	
	result->tag_name = strdup(elem);
	if(!result->tag_name) {
		error("Out of memory.");
		free(result);
		return NULL;
	}
	free(elem);

	/* Grag the arguments */
	args = get_rest(string);
	free_string(string);

	result->args = create_tag_args(args);

	return result;
}

/*
 * free_html_tag() 
 *
 * free's an html tag structure
 */

void free_html_tag(html_tag *tag)
{
	free_html_tag_args(tag->args);
	free(tag->tag_name);
	free(tag);
}

/*
 * get_next_tag()
 *
 * returns the next tag
 */

html_tag *get_next_tag()
{
	int c, counter;
	char *buf, *tmp;
	html_tag *result;
	unsigned int bufsize = HTML_TAG_BUFFER_SIZE;
		
	/* Skip ahead until next tag */

	if(find_next_tag() < 0)
		return NULL;
	
	/* Allocate a default buffer */

	buf = (char *)malloc(sizeof(char) * bufsize);
	if(!buf) {
		error("Not enough memory.");
		return NULL;
	}

	/* Retrieve the tag */

	counter = 0;

	while((c = buffer_get_char(source)) != '>') {

		/* Resize the buffer if necessary */

		if(counter >= bufsize) {

			tmp = strdup(buf);
			if(!tmp) {
				error("Not enough memory.\n");
				free(buf);
				return NULL;
			}

			bufsize += HTML_TAG_BUFFER_SIZE;

			buf = (char *)realloc(buf, sizeof(char) * bufsize);
			if(!buf) {
				error("Not enough memory.\n");
				free(tmp);
				return NULL;
			}

			strcpy(buf, tmp);
			free(tmp);
		}

		buf[counter++] = toupper(c);
	}
	buf[counter] = 0;

	/* Extract only the tag from buf */

	result = create_html_tag_struct(buf);
	free(buf);

	return result;
}

/*
 * html_parse()
 *
 * Let the parsing begin
 */

int html_parse_document(buffer_t *document)
{
	html_tag *tag;

	if(!document)
		return -1;

	if(html_parser_open(document) < 0)
		return -1;

	END_HTML_DOC = 0;

	while(!END_HTML_DOC) {
			
		tag = get_next_tag();
		if(!tag)
			break;
	
		process_tag(tag);

		free_html_tag(tag);
	}

	html_parser_close();

	return 0;
}

/*
 * process_tag()
 *
 * Process the html tags. 
 */

void process_tag(html_tag *tag)
{
	html_token *token;

	if(!tag)
		return;

	/* Lookup the tag and execute it's function */

	token = get_tag(tag->tag_name);

	if(token->tag_func) 
		token->tag_func(tag->args);

	return;
}
